--****************************************************************************
-- file:x_cylinder.ms
-- purpose:x junction cylinder proof of concept
-- author: klunk

plugin simpleObject xcylinder
	name:"xcylinder"
	classID:#(0x687c503c, 0x21d67a6b)
	category:"Scripted"
(
	parameters main rollout:params
	(	
		height		type:#float		default:25.0	ui:ui_height;
		radius		type:#float		default:25.0 	ui:ui_radius;
		sides		type:#integer	default:16		ui:ui_sides;
		cang		type:#float		default:0.0 	ui:ui_cang;
		rang		type:#float		default:90.0 	ui:ui_rang;
		lang		type:#float		default:-90.0 	ui:ui_lang;
		hedges 		type:#boolean	default:false	ui:ui_hedges;
		tedges		type:#boolean	default:false	ui:ui_tedges;
		edgweight 	type:#float		default:1.0 	ui:ui_edgweight;
	
		on rang set val do
		(
			if val < lang + 30 then 
				rang = lang + 30;
		)

		on lang set val do
		(
			if val > rang - 30 then 
				lang = rang - 30;
		)	
	)
	
--**************************************************************************	
	
	rollout params "Params"
	(
		spinner ui_height  "Height:" 	range:[0, 1e9, 0] fieldwidth:64 type:#float align:#right;
		spinner ui_radius  "Radius:" 	range:[0, 1e9, 0] fieldwidth:64 type:#float align:#right;
		spinner ui_sides   "Sides:" 	range:[3, 128, 16] fieldwidth:64 type:#integer align:#right;
		spinner ui_cang 	"Center Angle:" range:[-150, 150, 0] fieldwidth:64 type:#float align:#right scale:1;
		spinner ui_rang 	"Right Angle:" range:[-150, 150, 0] fieldwidth:64 type:#float align:#right scale:1;
		spinner ui_lang  	"Left Angle:" range:[-150, 150, 0] fieldwidth:64 type:#float align:#right scale:1;
		
		spinner ui_edgweight  	"Edge Weight:" range:[0,1024, 0] fieldwidth:64 type:#float align:#right scale:0.01;
		
		checkbox ui_hedges "Hard Edges" align:#left;
		checkbox ui_tedges "Turn Edges" align:#left;
	)
	
--**************************************************************************	
	
	tool create
	(
 		on mousePoint click do
		case click of
		(
			1:
			(
				nodeTM.translation = gridPoint;
				height = radius = 0.0;
			)	
			3: #stop
		)

		on mouseMove click do
		case click of
		(
			2: radius = amax (abs gridDist.y) (abs gridDist.x);
			3: height = gridDist.z
		)
	)
		
--**************************************************************************		

	fn gen_elipse a b sides tm offset =
	(	
		local incr = 360.0/sides;
		for s = 1 to sides collect
		(
			u = incr * (s - 1) + offset;
			[a * cos u, b * sin u, 0] * tm;
		)
	)	
		
--**************************************************************************	
	
	fn get_intersection_plane r1 r2 flipy alt =
	(
		zaxis = normalize (r1.dir + r2.dir)
		yaxis = if flipy then [0,1,0] else [0,-1,0] ;
		xaxis = normalize (cross yaxis zaxis);
		
		if alt then
			matrix3 zaxis -yaxis xaxis r2.pos;
		else
			matrix3 xaxis yaxis zaxis r2.pos;
	)
	
--**************************************************************************	
-- build an array of verts that makes up the 3 cylinder intersection	
	
	fn build_intersection_verts sray rray cray lray nsides even =
	(
-- stem to right arm intersection		
		
		tm = get_intersection_plane sray rray true false;
		ecc = dot sray.dir tm.row3;
		e1 = gen_elipse (radius/ecc) radius  nsides tm 90;
	
-- stem to left arm intersection		
		
		tm = get_intersection_plane sray lray false false;
		ecc = dot sray.dir tm.row3;
		e2 = gen_elipse  (radius/ecc) radius nsides tm -90;
		
-- right arm to center arm intersection			
		
		tm = get_intersection_plane rray cray false true;
		ecc = dot rray.dir tm.row3;
		e3 = gen_elipse  (radius/ecc) radius nsides tm 90;
		
-- center arrn to left		
		
		tm = get_intersection_plane cray lray false true;
		ecc = dot cray.dir tm.row3;
		e4 = gen_elipse  (radius/ecc) radius nsides tm 90;
		
-- collect the verts....		
		
		hsides = nsides/2;
		start = if even then hsides + 1 else hsides + 2;
		iverts = for v = 1 to start - 1 collect e2[v];
		join iverts (for v = start to nsides collect e1[v]);
		join iverts (for v = 2 to start - 1 collect e3[v]);
		join iverts (for v = 2 to start - 1 collect e4[v]);
		iverts
	)	

--**************************************************************************	
-- create indexing for the intersection verts as a ring of cylinder verts from below	

	fn get_in_vindices offset nsides =  (  for i = 1 to nsides collect i + offset; )
	
--**************************************************************************	
-- create indexing for the intersection verts as a ring of cylinder verts going to the right hand arm	
	
	fn get_out_right_vindices offset nsides even =
	(
		hnsides = nsides/2;
		end = nsides + hnsides;
		if even then end -= 1;
		start = if even then hnsides + 1 else hnsides + 2;
			
		indices = #(offset + 1);
		join indices (for i = nsides + 1 to end collect i + offset);
		join indices (for i = start to nsides collect i + offset);
	)	
	
--**************************************************************************	
-- create indexing for the intersection verts as a ring of cylinder verts going to the center   arm		
		
	fn get_out_center_vindices offset nsides even =
	(		
		hnsides = nsides/2;
		start = nsides + hnsides;
		if even then start -= 1;
		indices = #(offset + 1);	
		join indices (for i = 2 to hnsides + 1  collect i + offset + start - 1);
		if even then indices[indices.count] = offset + hnsides + 1;
		join indices (for i = start to nsides by - 1 collect i + offset);
	)		
	
--**************************************************************************	
-- create indexing for the intersection verts as a ring of cylinder verts going to the left hand arm		
		
	fn get_out_left_vindices offset nsides even =
	(		
		hnsides = nsides/2;
		start = nsides + nsides - 1;
		if even then start -= 1;
		indices = for i = 1 to hnsides + 1 collect i + offset;
		join indices (for i = start to nsides by - 1 collect i + offset);
	)	
	
--**************************************************************************

	fn set_face msh f verts vis matid smg =
	(
		setface msh f verts;
		setEdgeVis msh f 1 (vis[1] == 1);
		setEdgeVis msh f 2 (vis[2] == 1);
		setEdgeVis msh f 3 (vis[3] == 1);
		setFaceMatID msh f matid;
		setFaceSmoothGroup msh f smg;	
		f+1;	
	)	
	
--**************************************************************************	
	
	fn set_quad msh f verts matid smg =
	(
		setface msh f verts[1] verts[4] verts[3];
		setEdgeVis msh f 1 false;
		setEdgeVis msh f 2 true;
		setEdgeVis msh f 3 true;
		setFaceMatID msh f matid;
		setFaceSmoothGroup msh f smg;
		f+=1
		setface msh f verts[1] verts[2] verts[4];
		setEdgeVis msh f 1 true;
		setEdgeVis msh f 2 true;
		setEdgeVis msh f 3 false;
		setFaceMatID msh f matid;
		setFaceSmoothGroup msh f smg;
		f+1;	
	)	
	
--**************************************************************************	
	
	fn build_mesh = 
	(
		setquad = meshop.setQuad;
		--setquad = set_quad;
		
		meshop.supportVertWeights mesh;
		meshop.resetVertWeights mesh;
		
		smg1 = smg2 = smg3 = smg4 = 1;
		if hedges then
		(	
			smg2 = 2;
			smg3 = 4;
			smg4 = 8;
		)	
		
		even = bit.and sides 1 == 0;
		hsides = sides/2;
		dsides = 2 * sides;

		tvec = [0,0,1];
		sray = ray [0,0,0] tvec;
		pivot_point = [0,0,0] -- + height * tvec;
		
-- right arrn xform		
		
		rtm = preRotateY (matrix3 1) rang;
		rvec = tvec * rtm;
		rray = ray pivot_point rvec;
		rtm.translation = pivot_point + rvec * height;
	
-- left arrn xform		
		
		ltm = preRotateY (matrix3 1) lang;
		lvec = tvec * ltm;
		lray = ray pivot_point lvec;
		ltm.translation = pivot_point + lvec * height;
		
-- center arm xform

		ctm = preRotateY (matrix3 1) cang;
		cvec = tvec * ctm;
		cray = ray pivot_point cvec;
		ctm.translation = pivot_point + cvec * height;
		
-- build the intersection verts		
		
		iverts = build_intersection_verts sray rray cray lray sides even;
		num_tran_verts = iverts.count; 
		
		mesh.NumVerts = num_tran_verts + sides * 4;

-- base verts

		vi = 1;
		incr = 360.0/sides;
		for s = 1 to sides do -- initial ring of cylinder verts
		(
			u = 90 + incr * (s - 1);
			setvert  mesh vi [radius * cos u, radius * sin u, -height];
			vi += 1;
		)
		
-- intersection	verts
		
		ioffset = vi - 1; -- cache our indexing offset
		for v = 1 to num_tran_verts do  -- add our intersection verts
		(
			setvert mesh vi iverts[v];
			vi += 1;
		)	
		meshop.setVertWeight mesh #{(ioffset+1)..(vi-1)} edgweight;
		
-- right arm verts		
		
		for s = 1 to sides do -- right side end
		(
			u = 90 + incr * (s - 1);
			setvert mesh vi ([radius * cos u, radius * sin u, 0] * rtm);
			vi += 1;
		)
	
-- center arm verts		
		
		for s = 1 to sides do -- right side end
		(
			u = 90 + incr * (s - 1);
			setvert mesh vi ([radius * cos u, radius * sin u, 0] * ctm);
			vi += 1;
		)		
	
-- left arm verts	
		
		for s = 1 to sides do -- left side end
		(
			u = 90 + incr * (s - 1);
			setvert mesh vi ([radius * cos u, radius * sin u, 0] * ltm);
			vi += 1;
		)
		
-- create the faces		
		
		nfaces = 8 * sides
		if not even then nfaces += 2;
		mesh.NumFaces = nfaces;
		
-- base faces	
		
		voffset = 0;
		indices = get_in_vindices ioffset sides;
		fi = 1;
		for j = 1 to sides do
		(	
			a = j + voffset;
			b = a + 1;
			if j == sides then b = voffset + 1; 
				
			c = indices[j];
			d = indices[j+1];
			if j == sides then d = indices[1];
			
			fi = setQuad mesh fi [a,b,c,d] 1 smg1 short_diagonal:tedges;
		)
		
-- right arm/branch faces		
		
		voffset = sides + num_tran_verts; -- vert offset to cyl verts
		indices = get_out_right_vindices ioffset sides even;
		for j = 1 to sides do
		(	
			a = indices[j];
			b = indices[j+1];
			if j == sides then b = indices[1];
			
			c = j + voffset;
			d = c + 1;
			if j == sides then d = voffset + 1; 
			
			fi = setQuad mesh fi [a,b,c,d] 1 smg2 short_diagonal:tedges;
		)	

-- center arm/branch faces		
		
		voffset = dsides + num_tran_verts; -- vert offset to cyl verts
		indices = get_out_center_vindices ioffset sides even;
		for j = 1 to sides do
		(	
			a = indices[j];
			b = indices[j+1];
			if j == sides then b = indices[1];
			
			c = j + voffset;
			d = c + 1;
			if j == sides then d = voffset + 1; 
			
			fi = setQuad mesh fi [a,b,c,d] 1 smg3 short_diagonal:tedges;
		) 
		
-- left arm/branch faces		
		
		voffset = sides + dsides + num_tran_verts; -- vert offset to cyl verts
		indices = get_out_left_vindices ioffset sides even;
		for j = 1 to sides do
		(	
			a = indices[j];
			b = indices[j+1];
			if j == sides then b = indices[1];
			
			c = j + voffset;
			d = c + 1;
			if j == sides then d = voffset + 1; 
			
			fi = setQuad mesh fi [a,b,c,d] 1 smg4 short_diagonal:tedges;
		) 		
		
-- fillet faces		

		if not even then
		(
			a = sides + hsides + 1;
			b = dsides + sides - 1;
			c = a + 1;
			d = dsides + hsides
			fi = setQuad mesh fi [c, d, a, b] 1 smg1 short_diagonal:tedges;
		)	
		update mesh;
	)
	
	on update do build_mesh();
	on buildmesh do build_mesh();
)

--**************************************************************************
--**************************************************************************